# #############################################################################
#
# Copyright (C) OMG Plc 2014.
# All rights reserved.  This software is protected by copyright
# law and international treaties.  No part of this software / document
# may be reproduced or distributed in any form or by any means,
# whether transiently or incidentally to some other use of this software,
# without the written permission of the copyright owner.
#
# #############################################################################

import os
import sys
import math


try:
    import FbxCommon
    from fbx import *
except:
    print "Failed to import the Autodesk Filmbox Python SDK. Please download and install."
    sys.exit(1)

try:
    import numpy
except:
    print "Failed to import numpy. Please download and install."
    sys.exit(1)
    
try:
    from PyCara.PyCaraExport import *
except :
    print "Failed to import PyCara. Please download and install" 
    sys.exit(1)



# #############################################################################
# Utility functions
 
def format_timecode(timecode):
    return "{:0>2}:{:0>2}:{:0>2}:{:0>2}+{:0>2}".format( timecode[0], timecode[1], timecode[2], timecode[3], int(timecode[4]*100) )

def average_point( Xs ):
    if not Xs:
        return None
    accum = numpy.array([0.,0.,0.])
    for X in Xs:
        accum += X
    return accum / float(len(Xs))

    
# Helper class to represent the timecode standard from Cara
class ETimecodeStandard:
    ETimecode24=0
    ETimecode24Drift=1
    ETimecode25=2
    ETimecode30=3
    ETimecode30Drift=4
    ETimecode30Drop=5
    ETimecodeStandardCount=6
    ETimecodeNone = 0xFFFFFFFF
    
timecode_standards = [
 "ETimecode24", "ETimecode24Drift", "ETimecode25", "ETimecode30", "ETimecode30Drift" , "ETimecode30Drop" 
]

timecode_standards2 = [
 "24Hz Film", "23.97Hz Film Drift", "25Hz PAL", "30Hz", "29.97Hz NTSC Drift" , "29.97Hz NTSC Drop" 
]
    
def timecode_standard_to_string( standard ):
    if standard >= ETimecodeStandard.ETimecodeStandardCount:
        raise "Unknown ETimecodeStandard"
    return timecode_standards2[standard]

fbx_time_modes = [
 "eDefaultMode", "eFrames120", "eFrames100", "eFrames60",
 "eFrames50", "eFrames48", "eFrames30", "eFrames30Drop",
 "eNTSCDropFrame", "eNTSCFullFrame", "ePAL", "eFrames24",
 "eFrames1000", "eFilmFullFrame", "eCustom", "eFrames96",
 "eFrames72", "eFrames59dot94"
]
    
def fbx_time_mode_to_string( mode ):
    if mode >= FbxTime.eModesCount:
        raise "Unknown FbxTime.EMode"
    return fbx_time_modes[mode]
    
def timecode_standard_to_fbx_time_mode( standard ):
    if standard == ETimecodeStandard.ETimecode24:
        return FbxTime.eFrames24
    elif standard == ETimecodeStandard.ETimecode24Drift:
        return FbxTime.eFilmFullFrame
    elif standard == ETimecodeStandard.ETimecode25:
        return FbxTime.ePAL
    elif standard == ETimecodeStandard.ETimecode30:
        return FbxTime.eFrames30
    elif standard == ETimecodeStandard.ETimecode30Drift:
        return FbxTime.eNTSCFullFrame 
    elif standard == ETimecodeStandard.ETimecode30Drop:
        return FbxTime.eNTSCDropFrame         
    raise Exception("Unknown standard")
    

# #############################################################################
# 
  
def add_cameras( cara_data, cameras_parent_node, pSdkManager, lAnimLayer, start_time, time_mode, verbose=False ):
    """Read cameras from the cara data and add as FBX nodes
    
    For each camera in the cara file, we add a camera node with attached camera attributes, parented to a 
    a node which will contain the animated transformation representing the camera pose. 
    
    Each camera_transform node will be parented to the supplied camera_parent_node.    
    """
    
    # generate some temporary data that will be used below
    delta_time = FbxTime()        
    R_about_x_180 = numpy.array( [[1,0,0], [0,-1,0],[0,0,-1]] )
    R_about_y_180 = numpy.array( [[0,0,-1],[0,1,0], [1,0,0]] )
    
    # for each camera in the cara data 
    
    for camera_index in xrange( cara_data.n_cameras() ):
    
        camera_transform_node = FbxNode.Create(pSdkManager, "camera_transform_{}".format(camera_index) )
   
        camera_node = FbxNode.Create(pSdkManager, "camera_{}".format( camera_index ) )

        camera = FbxCamera.Create( pSdkManager, "camera_{}_attribute".format( camera_index ) )  

        # compute the camera attributes for this cameras intrinisic (internal) parameters
   
        image_size = cara_data.image_size(camera_index)

        radial_params, radial_centre = cara_data.radial_distortion(camera_index)
        radial_parameters = [radial_params[0], radial_params[1], radial_centre[0], radial_centre[1]] 

        left_top = vicon_radial_correct( radial_parameters, [0., 0.] )
        right_top = vicon_radial_correct( radial_parameters, [image_size[0], 0] )
        left_bottom = vicon_radial_correct( radial_parameters, [0., 720.] )
        right_bottom = vicon_radial_correct( radial_parameters, [image_size[0], 720.] )
       
        max_image_width = max( right_bottom[0], right_top[0] ) - min( left_bottom[0], left_top[0] )           
        corrected_image_width = max_image_width
                                
        camera.SetFormat( FbxCamera.eCustomFormat  )
        camera.SetAspect( FbxCamera.eFixedResolution, image_size[0], image_size[1] )
        camera.SetPixelRatio( 1.0 )
        camera.SetApertureMode(FbxCamera.eHorizontal )
        
        aperture_width = 1.0
        aperture_height = 1.0 * (image_size[1]/float(image_size[0]))
        camera.SetApertureWidth( aperture_width )
        camera.SetApertureHeight( aperture_height )  
        camera.SetViewFrustumBackPlaneMode( FbxCamera.ePlanesWhenMedia  )
        
        camera.NearPlane.Set(0.001)
        camera.FarPlane.Set(100000)
        
        focal_length_x_pixels = cara_data.focal_length(camera_index)

        # compute the camera's FOV using the radially corrected image width and the focal length in pixels
        fov = 2.0 * math.atan( (corrected_image_width*0.5) / focal_length_x_pixels) * (180.0 / math.pi)
   
        if verbose:
            print("camera_{} focal_length_x_pixels={}".format( camera_index, focal_length_x_pixels ))
            print("camera_{} corrected_image_width={}".format( camera_index, corrected_image_width ))
            print("camera_{} fov={:.3f}".format( camera_index, fov ))
        
        camera.FieldOfView.Set(fov);

        # setup the camera nodes
        
        camera_node.SetNodeAttribute( camera ) 
        cameras_parent_node.AddChild(camera_transform_node)
        camera_transform_node.AddChild(camera_node)

        # setup the animation curves for the camera pose
        
        x_curve = camera_transform_node.LclTranslation.GetCurve(lAnimLayer, "X", True)
        y_curve = camera_transform_node.LclTranslation.GetCurve(lAnimLayer, "Y", True)
        z_curve = camera_transform_node.LclTranslation.GetCurve(lAnimLayer, "Z", True)

        rx_curve = camera_transform_node.LclRotation.GetCurve(lAnimLayer, "X", True)
        ry_curve = camera_transform_node.LclRotation.GetCurve(lAnimLayer, "Y", True)
        rz_curve = camera_transform_node.LclRotation.GetCurve(lAnimLayer, "Z", True)
        
        x_curve.KeyModifyBegin()
        y_curve.KeyModifyBegin()
        z_curve.KeyModifyBegin()

        rx_curve.KeyModifyBegin()
        ry_curve.KeyModifyBegin()
        rz_curve.KeyModifyBegin()

        
        # for each frame in the cara data 
        
        for frame_index in xrange(cara_data.n_frames()):                

            # compute the FbxTime object for the current frame
            delta_time.SetFrame(frame_index, time_mode)
            lTime = delta_time + start_time

            # obtain the camera pose as a quaternion rotation and a translation
            (R_camera_quat, X) = cara_data.camera_pose(camera_index, frame_index) 
            if R_camera_quat is None:
                continue
            
            R_camera = quaternion_to_matrix( R_camera_quat )
            
            R_fbx =  numpy.dot( numpy.dot(R_about_y_180, R_about_x_180), numpy.array(R_camera) )
            R = numpy.linalg.inv( R_fbx )
              
            rotation_quat = matrix_to_quaternion( R )
                            
            fbx_euler_xyz = numpy.array(quaternion_to_euler_static_xyz( rotation_quat )) * (180./math.pi)
                            
            rx_key_index = rx_curve.KeyAdd(lTime)[0]
            rx_curve.KeySetValue(rx_key_index, fbx_euler_xyz[0])
            rx_curve.KeySetInterpolation(rx_key_index, FbxAnimCurveDef.eInterpolationCubic)
        
            ry_key_index = ry_curve.KeyAdd(lTime)[0]
            ry_curve.KeySetValue(ry_key_index, fbx_euler_xyz[1])
            ry_curve.KeySetInterpolation(ry_key_index, FbxAnimCurveDef.eInterpolationCubic)

            rz_key_index = rz_curve.KeyAdd(lTime)[0]
            rz_curve.KeySetValue(rz_key_index, fbx_euler_xyz[2])
            rz_curve.KeySetInterpolation(rz_key_index, FbxAnimCurveDef.eInterpolationCubic)
            
            x_key_index = x_curve.KeyAdd(lTime)[0]
            x_curve.KeySetValue(x_key_index, X[0])
            x_curve.KeySetInterpolation(x_key_index, FbxAnimCurveDef.eInterpolationCubic)
        
            y_key_index = y_curve.KeyAdd(lTime)[0]
            y_curve.KeySetValue(y_key_index, X[1])
            y_curve.KeySetInterpolation(y_key_index, FbxAnimCurveDef.eInterpolationCubic)

            z_key_index = z_curve.KeyAdd(lTime)[0]
            z_curve.KeySetValue(z_key_index, X[2])
            z_curve.KeySetInterpolation(z_key_index, FbxAnimCurveDef.eInterpolationCubic)

        x_curve.KeyModifyEnd()
        y_curve.KeyModifyEnd()
        z_curve.KeyModifyEnd()        

        rx_curve.KeyModifyEnd()
        ry_curve.KeyModifyEnd()
        rz_curve.KeyModifyEnd()        

def transform_data(cara_data, top_node, pSdkManager):
    """Translate so that the centroid of the face points is origin
    
    Read face points from the cara data at frame 0, compute their centroid and set that as the origin of the top_node.
    """

    face_centroid = None
    Xs = []
    for point_id in cara_data.face_points():
        X = cara_data.point_position(point_id, 0)
        if X is not None:
            Xs.append(X)

    face_centroid = average_point(Xs)

    if face_centroid is not None:
        top_node.LclTranslation.Set( FbxDouble3( -face_centroid[0], -face_centroid[1], -face_centroid[2] ) )
        top_node.SetRotationPivot( FbxNode.eSourcePivot, FbxVector4( face_centroid[0], face_centroid[1], face_centroid[2] ) )

     
def add_brim_markers(cara_data, brim_markers_parent_node, pSdkManager, verbose):
    """Read brim points from the cara data and add as FBX nodes"""
  

    count = 0
    for point_id in cara_data.brim_points():

        point_name = cara_data.point_name(point_id) 
        if point_name is "":
            point_name = "brim_marker_{}".format( count )
            count += 1
        if verbose:
            print("adding brim point [{}]".format( point_name ))

        X = cara_data.point_position(point_id, 0)
        if X is not None:        
            marker_node = FbxNode.Create(pSdkManager, point_name )
            marker = FbxMarker.Create( pSdkManager, "{}_attribute".format(point_name) )
            marker_node.SetNodeAttribute( marker )
            marker_node.LclTranslation.Set( FbxDouble3( X[0], X[1], X[2] ) )
            brim_markers_parent_node.AddChild(marker_node)

def add_face_markers(cara_data, face_markers_parent_node, pSdkManager, lAnimLayer, start_time, time_mode, verbose=False):
    """Read face points from the cara data and add as FBX nodes"""
  
    count = 0
    delta_time = FbxTime()
    for point_id in cara_data.face_points():

        point_name = cara_data.point_name(point_id) 
        if point_name is "":
            point_name = "face_marker_{}".format( count )
            count += 1
        if verbose:
            print("adding face point [{}]".format( point_name ))

        marker_node = FbxNode.Create(pSdkManager, point_name )
        marker = FbxMarker.Create( pSdkManager, "{}_attribute".format(point_name) )
        marker_node.SetNodeAttribute( marker )
        face_markers_parent_node.AddChild(marker_node)
                 
        # setup the animation curves for the point position
                    
        x_curve = marker_node.LclTranslation.GetCurve(lAnimLayer, "X", True)
        y_curve = marker_node.LclTranslation.GetCurve(lAnimLayer, "Y", True)
        z_curve = marker_node.LclTranslation.GetCurve(lAnimLayer, "Z", True)

        x_curve.KeyModifyBegin()
        y_curve.KeyModifyBegin()
        z_curve.KeyModifyBegin()
        
        point_positions = cara_data.frames_and_3d_for_point( point_id ) 

        if verbose and not point_positions:
            print("point {} had no 3d data".format( point_id ))

        for frame_index, X in point_positions.iteritems():                
        
            # compute the FbxTime object for the current frame
            delta_time.SetFrame(frame_index, time_mode)
            lTime = delta_time + start_time
            
            x_key_index = x_curve.KeyAdd(lTime)[0]
            x_curve.KeySetValue(x_key_index, X[0])
            x_curve.KeySetInterpolation(x_key_index, FbxAnimCurveDef.eInterpolationCubic)
        
            y_key_index = y_curve.KeyAdd(lTime)[0]
            y_curve.KeySetValue(y_key_index, X[1])
            y_curve.KeySetInterpolation(y_key_index, FbxAnimCurveDef.eInterpolationCubic)

            z_key_index = z_curve.KeyAdd(lTime)[0]
            z_curve.KeySetValue(z_key_index, X[2])
            z_curve.KeySetInterpolation(z_key_index, FbxAnimCurveDef.eInterpolationCubic)

        x_curve.KeyModifyEnd()
        y_curve.KeyModifyEnd()
        z_curve.KeyModifyEnd()
        
def create_scene(cara_data, pSdkManager, pScene, verbose = False, 
                 export_brim_markers = True, export_face_markers = True, export_cameras = True):
    # Create scene info
    lSceneInfo = FbxDocumentInfo.Create(pSdkManager, "SceneInfo")
    lSceneInfo.mTitle = "CaraPost FBX Export"
    lSceneInfo.mSubject = "Export from CaraPost."
    lSceneInfo.mAuthor = "CaraPost"
    lSceneInfo.mRevision = "rev. 1.0"
    lSceneInfo.mKeywords = "face cara"
    lSceneInfo.mComment = "Vicon Motion Systems Ltd."
    pScene.SetSceneInfo(lSceneInfo)
    
    global_settings = pScene.GetGlobalSettings()
    global_settings.SetSystemUnit( FbxSystemUnit.mm )
        
    top_node = FbxNode.Create(pSdkManager, "cara_root")
    pScene.GetRootNode().AddChild(top_node)

    # ##########################################
    # Setup the animation layer
    frame_rate = cara_data.frame_rate()
    if frame_rate is None:
        raise Exception("Unable to read the frame rate from the cara file.\nPlease ensure that the pico files are located in the same directory as the cara file \nor that the cara file contains the correct locations.")
        
    timecode_standard = cara_data.timecode_standard()
    if timecode_standard == ETimecodeStandard.ETimecodeNone:
        raise Exception("Unable to read the timecode standard from the cara file.\nPlease ensure that the pico files are located in the same directory as the cara file \nor that the cara file contains the correct locations.")
    
    end_frame = cara_data.n_frames()-1
    start_timecode = cara_data.timecode_from_index(0)
    end_timecode = cara_data.timecode_from_index( end_frame )
    
    # making the distinction here between internal and external time modes
    # this allows us to support setting the timecode using frames when constructing
    # the scene, and to output the frames with the correct SMPTE timecode format 
    #     
    internal_time_mode = FbxTime().ConvertFrameRateToTimeMode( frame_rate )
    external_time_mode = timecode_standard_to_fbx_time_mode(timecode_standard)


    if verbose:
        print("frame rate : {}Hz".format( frame_rate ) )
        print("timecode standard : {}".format( timecode_standard_to_string( timecode_standard) ) )
        print("frame {:>6} : {}".format( 0, format_timecode(start_timecode) ) )
        print("frame {:>6} : {}".format( end_frame, format_timecode(end_timecode) ) )
        print("internal FBX time mode : {}" .format( fbx_time_mode_to_string( internal_time_mode ) ) )
        print("external FBX time mode : {}" .format( fbx_time_mode_to_string( external_time_mode ) ) )   
    
    # First animation stack.
    lAnimStackName = "cara_animation_stack"
    lAnimStack = FbxAnimStack.Create(pScene, lAnimStackName)

    start_time = FbxTime()
    end_time = FbxTime()
    
    start_time.SetTime( start_timecode[0], start_timecode[1], 
                        start_timecode[2], start_timecode[3],
                        0, int(start_timecode[4]*100), external_time_mode )

    end_time.SetTime( end_timecode[0], end_timecode[1], 
                      end_timecode[2], end_timecode[3],
                      0, int(end_timecode[4]*100), external_time_mode )
    
    time_span = FbxTimeSpan( start_time, end_time )
    
    lAnimStack.SetReferenceTimeSpan( time_span )
    lAnimStack.SetLocalTimeSpan( time_span )

    # this controls how the frames will be interpreted by the application 
    # importing the FBX file. e.g. FbxTime.ePAL is PAL timecode
    global_settings.SetTimeMode( external_time_mode )
    global_settings.SetTimelineDefaultTimeSpan(time_span)
       
    # The animation nodes can only exist on AnimLayers therefore it is mandatory to
    # add at least one AnimLayer to the AnimStack. And for the purpose of this example,
    # one layer is all we need.
    lAnimLayer = FbxAnimLayer.Create(pScene, "cara_animation_layer")
    lAnimStack.AddMember(lAnimLayer)

        
    
    # ##########################################
    # Center the data

    transform_data(cara_data, top_node, pSdkManager)
    
    # ##########################################
    # Add the brim markers
    
    if export_brim_markers:
        brim_markers_parent_node = FbxNode.Create(pSdkManager, "brim_markers")
        top_node.AddChild(brim_markers_parent_node)

        add_brim_markers(cara_data, brim_markers_parent_node, pSdkManager, verbose)
        

    
    # ##########################################
    # Add the face markers
    if export_face_markers:
        face_markers_parent_node = FbxNode.Create(pSdkManager, "face_markers")
        top_node.AddChild(face_markers_parent_node)

        add_face_markers(cara_data, face_markers_parent_node, pSdkManager, lAnimLayer, start_time, internal_time_mode, verbose)
    
    # ##########################################
    # Add the cameras 

    if export_cameras:    
        cameras_parent_node = FbxNode.Create(pSdkManager, "cameras")
        top_node.AddChild(cameras_parent_node)

        add_cameras( cara_data, cameras_parent_node, pSdkManager, lAnimLayer, start_time, internal_time_mode, verbose)
  

        
def export_fbx( cara_filename, fbx_filename, verbose=False):
    
    if verbose:
        print("export_fbx({},{})".format( cara_filename, fbx_filename))
    
    # Prepare the FBX SDK.
    (lSdkManager, lScene) = FbxCommon.InitializeSdkObjects()

    # Open a cara file for reaading
    cara_data = CaraExport()

    ok = cara_data.open( cara_filename )
    if verbose:
        print("reading [{}] : {}".format( cara_filename, ok ))

    if not ok:
        print("Unable to read the cara file [{}]".format(cara_filename))
        lSdkManager.Destroy()
        sys.exit(1)

    # Generate an output filename if necessary
    if fbx_filename is None:
        (head,tail) = os.path.split(cara_filename)
        (root,ext) = os.path.splitext(tail)
        fbx_filename = root + '.fbx'

    # Create the scene.
    lResult = create_scene(cara_data, lSdkManager, lScene, verbose)

    if lResult is False:
        print("An error occurred while creating the FBX scene\n")
        lSdkManager.Destroy()
        sys.exit(1)

    # Save the scene to a file.  
    # format=0 for binary output
    format = 0
    lResult = FbxCommon.SaveScene(lSdkManager, lScene, fbx_filename, format)

    if lResult is False:
        print("An error occurred while saving the FBX scene")
        lSdkManager.Destroy()
        sys.exit(1)

    # Destroy all objects created by the FBX SDK.
    lSdkManager.Destroy()

    print("Successfully generated the FBX file [{}]".format( fbx_filename ))
    
    sys.exit(0)

    
if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument( 'input', nargs=1 )
    parser.add_argument( '--output', '-o' )
    parser.add_argument( '--verbose', '-v', action='store_true' )
    args = parser.parse_args()

    if args.verbose is False:
        from PyCara.PyDebugServices import set_global_log_level
        set_global_log_level(2) 
        
    export_fbx( cara_filename=args.input[0], fbx_filename=args.output, verbose=args.verbose)
